package dynclass;

import java.lang.reflect.InvocationHandler;
import java.util.HashMap;
import java.util.Map;

/**
 * ClassCreator turns a ClassDefinition into a real-live Class.
 * It also has utility methods for constructing instances of
 * classes it creates.
 */
public class ClassCreator {

    private static final Map<ClassLoader,DynClassLoader> classLoaders=new HashMap<ClassLoader, DynClassLoader>();

    /**
     * Create a new instance of the given class, which must have been
     * created by ClassCreator.  Method calls will be dispatched to
     * the given InvocationHandler.
     */
    public static Object newInstance(Class cls, InvocationHandler impl) throws Exception {

        if (cls == null) {
            throw new IllegalArgumentException("Class is null");
        }
        if (impl == null) {
            throw new IllegalArgumentException("Handler is null");
        }
        ClassLoader cl = cls.getClassLoader();
        if (!(cl instanceof DynClassLoader)) {
            throw new IllegalArgumentException("Not a DynClass: " + cls.getName());
        }

        return ((DynClassLoader)cl).instantiate(cls, impl);
    }

    /*
    public static Object newInstance(Class cls, 
                                     Object[] args, 
                                     InvocationHandler impl) throws Exception {
    }
    */

    /**
     * Create a Class from the given description.  Recent definitions are
     * cached, so if an equivalent ClassDefinition was provided recently
     * then there is a good chance the same class will be returned.
     * The class will be defined with a parent class loader of Thread.currentThread().getContextClassLoader()
     */
    public static Class defineClass(ClassDefinition def) {
        return defineClass(Thread.currentThread().getContextClassLoader(),def);
    }

    /**
     * Create a Class from the given description.  Recent definitions are
     * cached, so if an equivalent ClassDefinition was provided recently
     * then there is a good chance the same class will be returned.
     * @param parentClassLoader the parent classLoader to use. Must not be null.
     */
    public static Class defineClass(ClassLoader parentClassLoader, ClassDefinition def) {
        if (parentClassLoader==null) {
            throw new IllegalArgumentException("parentClassLoader must not be null");
        }
        DynClassLoader dynClassLoader;
        synchronized(classLoaders) {
            // We store a map of DynClassLoader for each parent ClassLoader, so we can reuse the previous DynClassLoader if possible
            dynClassLoader = classLoaders.get(parentClassLoader);
            if (dynClassLoader==null || dynClassLoader.getMethodCount() > 300) {
                dynClassLoader=new DynClassLoader(parentClassLoader);
                classLoaders.put(parentClassLoader,dynClassLoader);
            }
        }
        return dynClassLoader.defineClass(def);
    }
}
